/* Qt Cascading Style Sheet (qss)


    background: palette(Mid);  # Old value of the QDockWidget

*/


QDockWidget::title {
    background: #B8B8B8;
    color: pallete(Link);
    text-align: center;
}

/*
QDockWidget::title:hover {
    background: red;
    border: 1px solid red;
    color: yellow;
}*/

BasePanel {
    margin: 5px;
}

InspectorSelectionPane {
    margin: 5px;
    margin-bottom: 0px;
}

QFrame#message_label {
    color: #FF0066;
}

QMainWindow::separator {
    background: black;
    width: 1px;
    height: 1px;
}

ToggleColumnTableWidget::item {
    margin: 3px;
}

/*
QMainWindow::separator {
    background: palette(Midlight);
}

QMainWindow::separator:hover {
    background: palette(Mid);
}*/
/*
DockWidgetTitleBar {
    background: yellow;
    border: 1px solid black;
}*/
